@@ -29,7 +29,7 @@
 // versions of these for the 'clipboard' selection, as Visual mode has no use
 // for them.
 
-#if defined(FEAT_CLIPBOARD) || defined(PROTO)
+#if defined(FEAT_CLIPBOARD)
 
 #if defined(FEAT_WAYLAND_CLIPBOARD)
 
@@ -128,12 +128,18 @@ static int clip_wl_own_selection(Clipboard_T *cbd);
 static void clip_wl_lose_selection(Clipboard_T *cbd);
 static void clip_wl_set_selection(Clipboard_T *cbd);
 
-# if defined(USE_SYSTEM) && defined(PROTO)
+# if defined(USE_SYSTEM)
 static bool clip_wl_owner_exists(Clipboard_T *cbd);
 # endif
 
 #endif // FEAT_WAYLAND_CLIPBOARD
 
+#ifdef FEAT_CLIPBOARD_PROVIDER
+static int clip_provider_is_available(Clipboard_T *cbd, char_u *provider);
+static void clip_provider_set_selection(Clipboard_T *cbd, char_u *provider);
+static void clip_provider_request_selection(Clipboard_T *cbd, char_u *provider);
+#endif
+
 /*
  * Selection stuff using Visual mode, for cutting and pasting text to other
  * windows.
@@ -172,7 +178,6 @@ skip:
     }
 }
 
-#ifdef FEAT_WAYLAND_CLIPBOARD
     static void
 clip_init_single(Clipboard_T *cb, int can_use)
 {
@@ -188,7 +193,6 @@ clip_init_single(Clipboard_T *cb, int can_use)
     cb->end.col    = 0;
     cb->state      = SELECT_CLEARED;
 }
-#endif
 
 /*
  * Check whether the VIsual area has changed, and if so try to become the owner
@@ -256,7 +260,9 @@ clip_gen_own_selection(Clipboard_T *cbd UNUSED)
     }
     else if (clipmethod == CLIPMETHOD_OTHER)
     {
-#if !defined(FEAT_XCLIPBOARD) && !defined(FEAT_WAYLAND_CLIPBOARD)
+#if (!defined(FEAT_XCLIPBOARD) && !defined(FEAT_WAYLAND_CLIPBOARD) \
+	&& !defined(FEAT_CLIPBOARD_PROVIDER)) \
+	|| (!(defined(UNIX) || defined(MACOS_X)) && defined(FEAT_CLIPBOARD_PROVIDER))
 	return clip_mch_own_selection(cbd);
 #endif
     }
@@ -323,7 +329,9 @@ clip_gen_lose_selection(Clipboard_T *cbd UNUSED)
     }
     else if (clipmethod == CLIPMETHOD_OTHER)
     {
-#if !defined(FEAT_XCLIPBOARD) && !defined(FEAT_WAYLAND_CLIPBOARD)
+#if (!defined(FEAT_XCLIPBOARD) && !defined(FEAT_WAYLAND_CLIPBOARD) \
+	&& !defined(FEAT_CLIPBOARD_PROVIDER)) \
+	|| (!(defined(UNIX) || defined(MACOS_X)) && defined(FEAT_CLIPBOARD_PROVIDER))
 	clip_mch_lose_selection(cbd);
 #endif
     }
@@ -1055,7 +1063,7 @@ clip_process_selection(
 #endif
 }
 
-# if defined(FEAT_GUI) || defined(PROTO)
+# if defined(FEAT_GUI)
 /*
  * Redraw part of the selection if character at "row,col" is inside of it.
  * Only used for the GUI.
@@ -1356,7 +1364,7 @@ clip_gen_set_selection(Clipboard_T *cbd)
     {
 #ifdef FEAT_GUI
 	if (gui.in_use)
-	clip_mch_set_selection(cbd);
+	    clip_mch_set_selection(cbd);
 #endif
     }
     else if (clipmethod == CLIPMETHOD_WAYLAND)
@@ -1373,8 +1381,16 @@ clip_gen_set_selection(Clipboard_T *cbd)
     }
     else if (clipmethod == CLIPMETHOD_OTHER)
     {
-#if !defined(FEAT_XCLIPBOARD) && !defined(FEAT_WAYLAND_CLIPBOARD)
+#if (!defined(FEAT_XCLIPBOARD) && !defined(FEAT_WAYLAND_CLIPBOARD) \
+	&& !defined(FEAT_CLIPBOARD_PROVIDER)) \
+	|| (!(defined(UNIX) || defined(MACOS_X)) && defined(FEAT_CLIPBOARD_PROVIDER))
 	clip_mch_set_selection(cbd);
+#endif
+    }
+    else if (clipmethod == CLIPMETHOD_PROVIDER)
+    {
+#ifdef FEAT_CLIPBOARD_PROVIDER
+	clip_provider_set_selection(cbd, clipprovider_name);
 #endif
     }
 }
@@ -1403,14 +1419,22 @@ clip_gen_request_selection(Clipboard_T *cbd UNUSED)
     }
     else if (clipmethod == CLIPMETHOD_OTHER)
     {
-#if !defined(FEAT_XCLIPBOARD) && !defined(FEAT_WAYLAND_CLIPBOARD)
+#if (!defined(FEAT_XCLIPBOARD) && !defined(FEAT_WAYLAND_CLIPBOARD) \
+	&& !defined(FEAT_CLIPBOARD_PROVIDER)) \
+	|| (!(defined(UNIX) || defined(MACOS_X)) && defined(FEAT_CLIPBOARD_PROVIDER))
 	clip_mch_request_selection(cbd);
+#endif
+    }
+    else if (clipmethod == CLIPMETHOD_PROVIDER)
+    {
+#ifdef FEAT_CLIPBOARD_PROVIDER
+	clip_provider_request_selection(cbd, clipprovider_name);
 #endif
     }
 }
 
 #if (defined(FEAT_X11) && defined(FEAT_XCLIPBOARD) && defined(USE_SYSTEM)) \
-	|| defined(PROTO)
+
     static int
 clip_x11_owner_exists(Clipboard_T *cbd)
 {
@@ -1418,8 +1442,8 @@ clip_x11_owner_exists(Clipboard_T *cbd)
 }
 #endif
 
-#if ((defined(FEAT_X11) || defined(FEAT_WAYLAND_CLIPBOARD)) \
-	&& defined(USE_SYSTEM)) || defined(PROTO)
+#if (defined(FEAT_X11) || defined(FEAT_WAYLAND_CLIPBOARD)) \
+	&& defined(USE_SYSTEM)
     int
 clip_gen_owner_exists(Clipboard_T *cbd UNUSED)
 {
@@ -1550,7 +1574,7 @@ did_set_clipboard(optset_T *args UNUSED)
  * Stuff for the X clipboard.  Shared between VMS and Unix.
  */
 
-#if defined(FEAT_XCLIPBOARD) || defined(FEAT_GUI_X11) || defined(PROTO)
+#if defined(FEAT_XCLIPBOARD) || defined(FEAT_GUI_X11)
 # include <X11/Xatom.h>
 # include <X11/Intrinsic.h>
 
@@ -1742,7 +1766,7 @@ clip_x11_notify_cb(Widget w UNUSED, Atom *sel_atom UNUSED, Atom *target UNUSED)
 /*
  * Property callback to get a timestamp for XtOwnSelection.
  */
-# if (defined(FEAT_X11) && defined(FEAT_XCLIPBOARD)) || defined(PROTO)
+# if defined(FEAT_X11) && defined(FEAT_XCLIPBOARD)
     static void
 clip_x11_timestamp_cb(
     Widget	w,
@@ -2037,8 +2061,7 @@ clip_x11_set_selection(Clipboard_T *cbd UNUSED)
 
 #endif
 
-#if defined(FEAT_XCLIPBOARD) || defined(FEAT_GUI_X11) \
-    || defined(FEAT_GUI_GTK) || defined(PROTO)
+#if defined(FEAT_XCLIPBOARD) || defined(FEAT_GUI_X11) || defined(FEAT_GUI_GTK)
 /*
  * Get the contents of the X CUT_BUFFER0 and put it in "cbd".
  */
@@ -2114,7 +2137,7 @@ yank_cut_buffer0(Display *dpy, Clipboard_T *cbd)
  * 'permanent' of the two), otherwise the PRIMARY one.
  * For now, use a hard-coded sanity limit of 1Mb of data.
  */
-#if (defined(FEAT_X11) && defined(FEAT_CLIPBOARD)) || defined(PROTO)
+#if defined(FEAT_X11) && defined(FEAT_CLIPBOARD)
     void
 x11_export_final_selection(void)
 {
@@ -2269,10 +2292,18 @@ clip_get_selection(Clipboard_T *cbd)
     }
     else if (!is_clipboard_needs_update())
     {
-	clip_free_selection(cbd);
+#ifdef FEAT_CLIPBOARD_PROVIDER
+	// We will choose if we want to the free the selection if using provider
+	if (clipmethod != CLIPMETHOD_PROVIDER)
+	    clip_free_selection(cbd);
+#endif
 
 	// Try to get selected text from another window
 	clip_gen_request_selection(cbd);
+#ifdef FEAT_CLIPBOARD_PROVIDER
+	if (clipmethod == CLIPMETHOD_PROVIDER)
+	    clip_access_type = CLIP_ACCESS_IMPLICIT;
+#endif
     }
 }
 
@@ -2436,7 +2467,7 @@ adjust_clip_reg(int *rp)
     if ((!clip_star.available && *rp == '*') ||
 	   (!clip_plus.available && *rp == '+'))
     {
-	msg_warn_missing_clipboard();
+	msg_warn_missing_clipboard(!clip_plus.available, !clip_star.available);
 	*rp = 0;
     }
 }
@@ -2534,6 +2565,7 @@ clip_wl_destroy_buffer_store(clip_wl_buffer_store_T *store)
 	wl_shm_pool_destroy(store->pool);
 
     close(store->fd);
+
     vim_free(store);
 }
 
@@ -2552,6 +2584,7 @@ clip_wl_init_buffer_store(int width, int height)
 	return NULL;
 
     store->available = false;
+
     store->width = width;
     store->height = height;
     store->stride = store->width * 4;
@@ -2827,6 +2860,7 @@ vwl_data_device_listener_event_selection(
     // Save offer. When we want to request data, then we'll actually call the
     // receive method.
     sel->offer = offer;
+
 }
 
     static void
@@ -2843,13 +2877,11 @@ vwl_data_device_listener_event_finished(
     {
 	sel = &clip_wl.regular;
 	clip_wl.regular.device = NULL;
-	clip_init_single(&clip_plus, FALSE);
     }
     else if (device == clip_wl.primary.device)
     {
 	sel = &clip_wl.primary;
 	clip_wl.primary.device = NULL;
-	clip_init_single(&clip_star, FALSE);
     }
     else
 	// Shouldn't happen
@@ -2963,14 +2995,6 @@ clip_init_wayland(void)
 	vwl_data_device_add_listener(clip_wl.primary.device,
 		&vwl_data_device_listener, NULL);
 
-    if (clipmethod == CLIPMETHOD_WAYLAND)
-    {
-	if (clip_wl.regular.available)
-	    clip_init_single(&clip_plus, TRUE);
-	if (clip_wl.primary.available)
-	    clip_init_single(&clip_star, TRUE);
-    }
-
     return OK;
 }
 
@@ -2979,6 +3003,14 @@ clip_uninit_wayland(void)
 {
     clip_wl_selection_T *sel;
 
+    if (clipmethod == CLIPMETHOD_WAYLAND)
+    {
+	if (clip_star.owned)
+	    clip_lose_selection(&clip_star);
+	if (clip_plus.owned)
+	    clip_lose_selection(&clip_plus);
+    }
+
 #ifdef FEAT_WAYLAND_CLIPBOARD_FS
     clip_wl_destroy_buffer_store(clip_wl.fs_buffer);
 #endif
@@ -3005,25 +3037,16 @@ clip_uninit_wayland(void)
 	sel = &clip_wl.primary;
     }
 
-    if (clipmethod == CLIPMETHOD_WAYLAND)
-	clip_init(FALSE);
-
     vim_memset(&clip_wl, 0, sizeof(clip_wl));
 }
 
     int
 clip_reset_wayland(void)
 {
-    if (clipmethod == CLIPMETHOD_WAYLAND)
-    {
-	if (clip_star.owned)
-	    clip_lose_selection(&clip_star);
-	if (clip_plus.owned)
-	    clip_lose_selection(&clip_plus);
-    }
-    clip_uninit_wayland();
+    wayland_uninit_connection();
 
-    if (clip_init_wayland() == FAIL)
+    if (wayland_init_connection(wayland_display_name) == FAIL
+	    || clip_init_wayland() == FAIL)
 	return FAIL;
 
     choose_clipmethod();
@@ -3426,7 +3449,7 @@ clip_wl_set_selection(Clipboard_T *cbd UNUSED)
 {
 }
 
-#if defined(USE_SYSTEM) && defined(PROTO)
+#if defined(USE_SYSTEM)
 /*
  * Return true if we own the selection corresponding to cbd or another client
  * does.
@@ -3451,7 +3474,7 @@ clip_wl_owner_exists(Clipboard_T *cbd)
  * depending on the order of values in str.
  */
     static clipmethod_T
-get_clipmethod(char_u *str)
+get_clipmethod(char_u *str, bool *plus UNUSED, bool *star)
 {
     int		len	= (int)STRLEN(str) + 1;
     char_u	*buf	= alloc(len);
@@ -3476,7 +3499,11 @@ get_clipmethod(char_u *str)
 	    {
 #ifdef FEAT_WAYLAND_CLIPBOARD
 		if (clip_wl.regular.available || clip_wl.primary.available)
+		{
 		    method = CLIPMETHOD_WAYLAND;
+		    *plus = clip_wl.regular.available;
+		    *star = clip_wl.primary.available;
+		}
 #endif
 	    }
 	}
@@ -3497,6 +3524,7 @@ get_clipmethod(char_u *str)
 		    // won't be executed since xterm_dpy will be set to NULL.
 		    xterm_update();
 		    method = CLIPMETHOD_X11;
+		    *plus = *star = true;
 		}
 #endif
 	    }
@@ -3505,21 +3533,57 @@ get_clipmethod(char_u *str)
 	{
 #ifdef FEAT_GUI
 	    if (gui.in_use)
+	    {
 		method = CLIPMETHOD_GUI;
+		*star = *plus = true;
+	    }
 #endif
 	}
 	else if (STRCMP(buf, "other") == 0)
 	{
-#if !defined(FEAT_XCLIPBOARD) && !defined(FEAT_WAYLAND_CLIPBOARD)
+#if (!defined(FEAT_XCLIPBOARD) && !defined(FEAT_WAYLAND_CLIPBOARD) \
+	&& !defined(FEAT_CLIPBOARD_PROVIDER)) \
+	|| (!(defined(UNIX) || defined(MACOS_X)) && defined(FEAT_CLIPBOARD_PROVIDER))
 		method = CLIPMETHOD_OTHER;
+		*plus = *star = true;
 #endif
 	}
 	else
 	{
-	    ret = CLIPMETHOD_FAIL;
-	    goto exit;
-	}
+#ifdef FEAT_CLIPBOARD_PROVIDER
+	    // Check if it is the name of a provider
+#ifndef ONE_CLIPBOARD
+	    int plus_avail = clip_provider_is_available(&clip_plus, buf);
+#endif
+	    int star_avail = clip_provider_is_available(&clip_star, buf);
 
+	    if (
+#ifndef ONE_CLIPBOARD
+		    plus_avail == 1 ||
+#endif
+		    star_avail == 1)
+	    {
+		method = CLIPMETHOD_PROVIDER;
+
+		vim_free(clipprovider_name);
+		clipprovider_name = vim_strsave(buf);
+
+#ifndef ONE_CLIPBOARD
+		*plus = plus_avail == 1;
+#endif
+		*star = star_avail == 1;
+	    }
+	    else if (
+#ifndef ONE_CLIPBOARD
+		    plus_avail == -1 ||
+#endif
+		    star_avail == -1)
+#endif
+	    {
+		ret = CLIPMETHOD_FAIL;
+		goto exit;
+	    }
+	}
 	// Keep on going in order to catch errors
 	if (method != CLIPMETHOD_NONE && ret == CLIPMETHOD_FAIL)
 	    ret = method;
@@ -3550,6 +3614,15 @@ clipmethod_to_str(clipmethod_T method)
 	    return (char_u *)"gui";
 	case CLIPMETHOD_OTHER:
 	    return (char_u *)"other";
+	case CLIPMETHOD_PROVIDER:
+#ifdef FEAT_CLIPBOARD_PROVIDER
+	    if (clipprovider_name == NULL)
+		return (char_u *)"none";
+	    else
+		return clipprovider_name;
+#else
+	    return (char_u *)"none";
+#endif
 	default:
 	    return (char_u *)"none";
     }
@@ -3562,9 +3635,9 @@ clipmethod_to_str(clipmethod_T method)
     char *
 choose_clipmethod(void)
 {
-    // We call get_clipmethod first so that we can catch any errors, even if
-    // clipmethod is useless
-    clipmethod_T method = get_clipmethod(p_cpm);
+    bool plus = false;
+    bool star = false;
+    clipmethod_T method = get_clipmethod(p_cpm, &plus, &star);
 
     if (method == CLIPMETHOD_FAIL)
 	return e_invalid_argument;
@@ -3580,20 +3653,37 @@ choose_clipmethod(void)
     if (method == CLIPMETHOD_NONE)
 	clip_init(FALSE);
     // If we have a clipmethod that works now, then initialize clipboard
-    else if (clipmethod == CLIPMETHOD_NONE
-	    && method != CLIPMETHOD_NONE)
+    else if (clipmethod == CLIPMETHOD_NONE && method != CLIPMETHOD_NONE)
     {
-	clip_init(TRUE);
-	did_warn_clipboard = FALSE;
+#ifndef ONE_CLIPBOARD
+	clip_init_single(&clip_plus, plus);
+	clip_plus.did_warn = false;
+#endif
+	clip_init_single(&clip_star, star);
+	clip_star.did_warn = false;
     }
-
-    // Disown clipboard if we are switching to a new method
-    if (clipmethod != CLIPMETHOD_NONE && method != clipmethod)
+    else if ((clipmethod != CLIPMETHOD_NONE && method != clipmethod))
     {
+	// Disown clipboard if we are switching to a new method
 	if (clip_star.owned)
 	    clip_lose_selection(&clip_star);
+	clip_init_single(&clip_star, star);
+#ifndef ONE_CLIPBOARD
 	if (clip_plus.owned)
 	    clip_lose_selection(&clip_plus);
+	clip_init_single(&clip_plus, plus);
+#endif
+    }
+    else
+    {
+	// If availability of a clipboard changed, then update the clipboard
+	// structure.
+#ifndef ONE_CLIPBOARD
+	if (plus != clip_plus.available)
+	    clip_init_single(&clip_plus, plus);
+#endif
+	if (star != clip_star.available)
+	    clip_init_single(&clip_star, star);
     }
 
     clipmethod = method;
@@ -3622,4 +3712,325 @@ ex_clipreset(exarg_T *eap UNUSED)
 		clipmethod_to_str(clipmethod));
 }
 
+#ifdef FEAT_CLIPBOARD_PROVIDER
+
+/*
+ * Check if a clipboard provider with given name exists and is available for the
+ * given clipboard. Returns 1 if the provider exists and the 'available'
+ * function returned true, 0 if the provider exists but the function returned
+ * false, and -1 on error.
+ */
+    static int
+clip_provider_is_available(Clipboard_T *cbd, char_u *provider)
+{
+    dict_T	*providers = get_vim_var_dict(VV_CLIPPROVIDERS);
+    typval_T	provider_tv = {0};
+    callback_T	callback = {0};
+    typval_T	rettv = {0};
+    typval_T	func_tv = {0};
+    char_u	*avail;
+    int		res = 0;
+
+    if (dict_get_tv(providers, (char *)provider, &provider_tv) == FAIL
+	    || provider_tv.v_type != VAR_DICT)
+	return -1;
+
+    if (dict_get_tv(provider_tv.vval.v_dict, "available", &func_tv) == FAIL)
+    {
+	clear_tv(&provider_tv);
+	// If "available" functon not specified assume always TRUE
+	return 1;
+    }
+
+    if ((callback = get_callback(&func_tv)).cb_name == NULL)
+	goto fail;
+
+    if (call_callback(&callback, -1, &rettv, 0, NULL) == FAIL ||
+	    rettv.v_type != VAR_STRING)
+	goto fail;
+
+    avail = rettv.vval.v_string;
+
+
+    if (
+#ifndef ONE_CLIPBOARD
+	    (vim_strchr(avail, '+') != NULL && cbd == &clip_plus) ||
+#endif
+	    (vim_strchr(avail, '*') != NULL && cbd == &clip_star))
+	res = 1;
+
+    if (FALSE)
+fail:
+	res = -1;
+
+    free_callback(&callback);
+    clear_tv(&func_tv);
+    clear_tv(&rettv);
+    clear_tv(&provider_tv);
+
+    return res;
+}
+
+/*
+ * Get the specified callback "function" from the provider dictionary of for
+ * register "reg".
+ */
+    static int
+clip_provider_get_callback(
+	char_u *reg,
+	char_u *provider,
+	char_u *function,
+	callback_T *callback)
+{
+    dict_T	*providers = get_vim_var_dict(VV_CLIPPROVIDERS);
+    typval_T	provider_tv;
+    typval_T	action_tv;
+    typval_T	func_tv;
+    callback_T	cb;
+
+    if (dict_get_tv(providers, (char *)provider, &provider_tv) == FAIL)
+	return FAIL;
+    else if (provider_tv.v_type != VAR_DICT)
+    {
+	clear_tv(&provider_tv);
+	return FAIL;
+    }
+    else if (dict_get_tv(
+		provider_tv.vval.v_dict,
+		(char *)function,
+		&action_tv) == FAIL)
+    {
+	clear_tv(&provider_tv);
+	return FAIL;
+    }
+    else if (action_tv.v_type != VAR_DICT)
+    {
+	clear_tv(&provider_tv);
+	clear_tv(&action_tv);
+	return FAIL;
+    }
+    else if (dict_get_tv(action_tv.vval.v_dict, (char *)reg, &func_tv) == FAIL)
+    {
+	clear_tv(&provider_tv);
+	clear_tv(&action_tv);
+	return FAIL;
+    }
+    else if ((cb = get_callback(&func_tv)).cb_name == NULL)
+    {
+	clear_tv(&provider_tv);
+	clear_tv(&action_tv);
+	clear_tv(&func_tv);
+	return FAIL;
+    }
+    clear_tv(&provider_tv);
+    clear_tv(&action_tv);
+
+    // func_tv owns the function name, so we must make a copy for the callback
+    set_callback(callback, &cb);
+    free_callback(&cb);
+    clear_tv(&func_tv);
+    return OK;
+}
+
+    static void
+clip_provider_set_selection(Clipboard_T *cbd, char_u *provider)
+{
+    char_u	*reg = (char_u *)(cbd == &clip_star ? "*" : "+");
+    callback_T	callback;
+    typval_T	rettv;
+    typval_T	argvars[4];
+    yankreg_T	*y_ptr;
+    char_u	type[2 + NUMBUFLEN] = {0};
+    list_T	*list = NULL;
+
+    if (clip_provider_get_callback(
+		reg,
+		provider,
+		(char_u *)"copy",
+		&callback) == FAIL)
+	return;
+
+    // Possibly get selected text, if using autoselect for 'clipboard'
+    cbd->owned = TRUE;
+    clip_get_selection(cbd);
+    cbd->owned = FALSE;
+
+    // Convert register type into a string
+    if (cbd == &clip_plus)
+	y_ptr = get_y_register(PLUS_REGISTER);
+    else
+	y_ptr = get_y_register(STAR_REGISTER);
+
+    switch (y_ptr->y_type)
+    {
+	case MCHAR:
+	    type[0] = 'v';
+	    break;
+	case MLINE:
+	    type[0] = 'V';
+	    break;
+	case MBLOCK:
+	    sprintf((char *)type, "%c%d", Ctrl_V, y_ptr->y_width + 1);
+	    break;
+	default:
+	    type[0] = 0;
+	    break;
+    }
+
+    argvars[0].v_type = VAR_STRING;
+    argvars[0].vval.v_string = reg;
+
+    argvars[1].v_type = VAR_STRING;
+    argvars[1].vval.v_string = type;
+
+    // Get register contents by creating a list of lines
+    list = list_alloc();
+
+    if (list == NULL)
+    {
+	free_callback(&callback);
+	return;
+    }
+
+    for (int i = 0; i < y_ptr->y_size; i++)
+	if (list_append_string(list, y_ptr->y_array[i].string, -1) == FAIL)
+	{
+	    free_callback(&callback);
+	    list_unref(list);
+	    return;
+	}
+
+    list->lv_refcount++;
+
+    argvars[2].v_type = VAR_LIST;
+    argvars[2].v_lock = VAR_FIXED;
+    argvars[2].vval.v_list = list;
+
+    argvars[3].v_type = VAR_UNKNOWN;
+
+    textlock++;
+    call_callback(&callback, -1, &rettv, 3, argvars);
+    clear_tv(&rettv);
+    textlock--;
+
+    free_callback(&callback);
+    list_unref(list);
+}
+
+    static void
+clip_provider_request_selection(Clipboard_T *cbd, char_u *provider)
+{
+    char_u	*reg = (char_u *)(cbd == &clip_star ? "*" : "+");
+    callback_T	callback;
+    typval_T	argvars[3];
+    typval_T	rettv;
+    int		ret;
+    char_u	*reg_type;
+    list_T	*lines;
+
+    if (clip_provider_get_callback(
+		reg,
+		provider,
+		(char_u *)"paste",
+		&callback) == FAIL)
+	return;
+
+    argvars[0].v_type = VAR_STRING;
+    argvars[0].vval.v_string = reg;
+
+    argvars[1].v_type = VAR_STRING;
+    argvars[1].vval.v_string = (char_u *)
+	(clip_access_type == CLIP_ACCESS_EXPLICIT ? "explicit" : "implicit");
+
+    argvars[2].v_type = VAR_UNKNOWN;
+
+    textlock++;
+    ret = call_callback(&callback, -1, &rettv, 2, argvars);
+    textlock--;
+
+    if (ret == FAIL)
+	goto exit;
+    else if (rettv.v_type == VAR_STRING
+	    && STRCMP(rettv.vval.v_string, "clear") == 0)
+    {
+	clip_free_selection(cbd);
+	goto exit;
+    }
+    else if (rettv.v_type == VAR_STRING
+	    && STRCMP(rettv.vval.v_string, "previous") == 0)
+	goto exit;
+    else if (rettv.v_type == VAR_TUPLE
+	    && TUPLE_LEN(rettv.vval.v_tuple) == 2
+	    && TUPLE_ITEM(rettv.vval.v_tuple, 0)->v_type == VAR_STRING
+	    && TUPLE_ITEM(rettv.vval.v_tuple, 1)->v_type == VAR_LIST)
+    {
+	reg_type = TUPLE_ITEM(rettv.vval.v_tuple, 0)->vval.v_string;
+	lines = TUPLE_ITEM(rettv.vval.v_tuple, 1)->vval.v_list;
+    }
+    else if (rettv.v_type == VAR_LIST
+	    && rettv.vval.v_list->lv_len == 2
+	    && rettv.vval.v_list->lv_first->li_tv.v_type == VAR_STRING
+	    && rettv.vval.v_list->lv_first->li_next->li_tv.v_type == VAR_LIST)
+    {
+	reg_type = rettv.vval.v_list->lv_first->li_tv.vval.v_string;
+	lines = rettv.vval.v_list->lv_first->li_next->li_tv.vval.v_list;
+    }
+    else
+	goto exit;
+
+    {
+	char_u	    yank_type = MAUTO;
+	long	    block_len = -1;
+	yankreg_T   *y_ptr;
+	char_u	    **contents;
+	listitem_T  *li;
+	int	    i = 0;
+
+	contents = ALLOC_MULT(char_u *, lines->lv_len + 1); // Ends with a NULL
+
+	if (contents == NULL)
+	    goto exit;
+
+	// Convert strings in list to type char_u **
+	FOR_ALL_LIST_ITEMS(lines, li)
+	{
+	    char_u *str = tv_get_string_chk(&li->li_tv);
+
+	    if (str == NULL)
+		goto exit;
+
+	    contents[i++] = vim_strsave(str);
+	}
+	contents[i] = NULL;
+
+	if (STRLEN(reg_type) > 0
+		&& get_yank_type(&reg_type, &yank_type, &block_len) == FAIL)
+	    goto exit;
+
+	if (cbd == &clip_plus)
+	    y_ptr = get_y_register(PLUS_REGISTER);
+	else
+	    y_ptr = get_y_register(STAR_REGISTER);
+
+	clip_free_selection(cbd);
+
+	str_to_reg(y_ptr,
+		yank_type,
+		(char_u *)contents,
+		STRLEN(contents),
+		block_len,
+		TRUE);
+
+	for (int k = 0; k < i; k++)
+	    vim_free(contents[k]);
+	vim_free(contents);
+    }
+
+exit:
+    free_callback(&callback);
+    clear_tv(&rettv);
+}
+
+#endif // FEAT_CLIPBOARD_PROVIDER
+
 #endif // FEAT_CLIPBOARD
