@@ -44,21 +44,51 @@ static int pum_win_width;
 // makes pum_visible() return FALSE even when there is a popup menu.
 static int pum_pretend_not_visible = FALSE;
 
+// Border support
+static int pum_border = 0;
+static int pum_margin = 0;	// margin of 1 cell on left and right
+static int pum_shadow = 0;
+
+// Border characters
+static struct {
+    int top;
+    int right;
+    int bottom;
+    int left;
+    int top_left;
+    int top_right;
+    int bottom_right;
+    int bottom_left;
+} pum_border_chars;
+
 static int pum_set_selected(int n, int repeat);
+static void pum_draw_border(void);
+static void pum_draw_shadow(void);
+static void compute_margins(int *right_margin, int *left_margin, int *left_padding);
 
 #define PUM_DEF_HEIGHT 10
 
+#define DEFINE_PUM_SETTER(feature) \
+	void \
+    pum_set_##feature(int enable) \
+    { \
+	pum_##feature = enable ? 1 : 0; \
+    }
+
+DEFINE_PUM_SETTER(border)
+DEFINE_PUM_SETTER(shadow)
+DEFINE_PUM_SETTER(margin)
+
     static void
 pum_compute_size(void)
 {
-    int	i;
     int	w;
 
     // Compute the width of the widest match and the widest extra.
     pum_base_width = 0;
     pum_kind_width = 0;
     pum_extra_width = 0;
-    for (i = 0; i < pum_size; ++i)
+    for (int i = 0; i < pum_size; ++i)
     {
 	if (pum_array[i].pum_text != NULL)
 	{
@@ -81,6 +111,175 @@ pum_compute_size(void)
     }
 }
 
+/*
+ * Calculate vertical placement for popup menu.
+ * Sets pum_row and pum_height based on available space.
+ */
+    static void
+pum_compute_vertical_placement(
+	int size,
+	int above_row,
+	int below_row)
+{
+    int context_lines;
+    int cline_visible_offset;
+    int extra_height = 2 * pum_border + pum_shadow;
+    int extra_offset = pum_border + pum_shadow;
+
+    pum_height = MIN(size, PUM_DEF_HEIGHT);
+    if (p_ph > 0 && pum_height > p_ph)
+	pum_height = p_ph;
+
+    // Put the pum below "pum_win_row" if possible.  If there are few lines
+    // decide on where there is more room.
+    if (pum_win_row + 2 >= below_row - pum_height - extra_height
+	    && pum_win_row - above_row > (below_row - above_row) / 2)
+    {
+	// pum above "pum_win_row"
+	if (State & MODE_CMDLINE)
+	    // for cmdline pum, no need for context lines
+	    context_lines = 0;
+	else
+	    // Leave two lines of context if possible
+	    context_lines = MIN(2, curwin->w_wrow - curwin->w_cline_row);
+
+	if (pum_win_row >= size + context_lines + extra_height)
+	{
+	    pum_row = pum_win_row - size - context_lines - extra_offset;
+	    pum_height = size;
+	}
+	else
+	{
+	    pum_row = pum_border;
+	    pum_height = pum_win_row - context_lines - extra_height;
+	}
+	if (p_ph > 0 && pum_height > p_ph)
+	{
+	    pum_row += pum_height - p_ph;
+	    pum_height = p_ph;
+	}
+    }
+    else
+    {
+	// pum below "pum_win_row"
+	if (State & MODE_CMDLINE)
+	    // for cmdline pum, no need for context lines
+	    context_lines = 0;
+	else
+	{
+	    // Leave three lines of context if possible
+	    validate_cheight();
+	    cline_visible_offset = curwin->w_cline_row +
+		curwin->w_cline_height - curwin->w_wrow;
+	    context_lines = MIN(3, cline_visible_offset);
+	}
+
+	pum_row = pum_win_row + context_lines + pum_border;
+	pum_height = MIN(below_row - pum_row - extra_offset, size);
+	if (p_ph > 0 && pum_height > p_ph)
+	    pum_height = p_ph;
+    }
+
+#if defined(FEAT_QUICKFIX)
+    // If there is a preview window above avoid drawing over it.
+    if (above_row > 0 && pum_row < above_row && pum_height > above_row)
+    {
+	pum_row = above_row;
+	pum_height = pum_win_row - above_row;
+    }
+#endif
+}
+
+/*
+ * Try to set 'pum_width' so that it fits within available_width.
+ * Returns TRUE if pum_width was successfully set, FALSE otherwise.
+ */
+    static int
+set_pum_width_aligned_with_cursor(int width, int available_width)
+{
+    int end_padding = TRUE;
+
+    if (width < p_pw)
+    {
+	width = p_pw;
+	end_padding = FALSE;
+    }
+    if (p_pmw > 0 && width >= p_pmw)
+    {
+	width = p_pmw;
+	end_padding = FALSE;
+    }
+
+    pum_width = width + (end_padding && width >= p_pw ? 1 : 0);
+    return available_width >= pum_width;
+}
+
+/*
+ * Calculate horizontal placement for popup menu. Sets pum_col and pum_width
+ * based on cursor position and available space.
+ */
+    static void
+pum_compute_horizontal_placement(int cursor_col)
+{
+    int desired_width = pum_base_width + pum_kind_width + pum_extra_width;
+    int available_width;
+    int col_offset;
+    int trailing_overhead = pum_scrollbar + pum_border
+	+ (pum_margin && pum_border ? 1 : 0) + (pum_shadow ? 2 : 0);
+
+#ifdef FEAT_RIGHTLEFT
+    if (pum_rl)
+    {
+	col_offset = pum_border && (cursor_col >= Columns - 2)
+	    ? (3 - Columns + cursor_col) : 0;
+	available_width = cursor_col + 1;
+	pum_col = cursor_col - col_offset;
+    }
+    else
+#endif
+    {
+	col_offset = pum_border && (cursor_col <= 1) ? (2 - cursor_col) : 0;
+	available_width = Columns - cursor_col;
+	pum_col = cursor_col + col_offset;
+    }
+
+    available_width -= trailing_overhead + col_offset;
+
+    // Case 1: Align pum with "cursor_col"
+    if (set_pum_width_aligned_with_cursor(desired_width, available_width))
+	return;
+    // Show the pum truncated, provided it is at least as wide as 'pum_width'
+    if (available_width > p_pw)
+    {
+	pum_width = available_width;
+	return;
+    }
+
+    // Case 2: Truncated pum is no longer aligned with "cursor_col"
+    available_width = Columns - trailing_overhead - col_offset;
+
+    if (available_width > p_pw)
+    {
+	pum_width = p_pw;  // Truncate beyond 'pum_width'
+#ifdef FEAT_RIGHTLEFT
+	if (pum_rl)
+	    pum_col = pum_width + trailing_overhead;
+	else
+#endif
+	    pum_col = Columns - pum_width - trailing_overhead;
+	return;
+    }
+
+    // Case 3: Not enough room anywhere, use what we have
+#ifdef FEAT_RIGHTLEFT
+    if (pum_rl)
+	pum_col = Columns - 1 - col_offset;
+    else
+#endif
+	pum_col = col_offset;
+    pum_width = Columns - trailing_overhead;
+}
+
 /*
  * Show the popup menu with items "array[size]".
  * "array" must remain valid until pum_undisplay() is called!
@@ -89,34 +288,25 @@ pum_compute_size(void)
  */
     void
 pum_display(
-    pumitem_T	*array,
-    int		size,
-    int		selected)	// index of initially selected item, none if
+	pumitem_T   *array,
+	int	    size,
+	int	    selected)   // index of initially selected item, -1 if
 				// out of range
 {
-    int		def_width;
-    int		max_width;
-    int		context_lines;
-    int		cursor_col;
-    int		above_row;
-    int		below_row;
-    int		cline_visible_offset;
-    int		content_width;
-    int		right_edge_col;
-    int		redo_count = 0;
+    int	    cursor_col;
+    int	    above_row;
+    int	    below_row;
+    int	    redo_count = 0;
 #if defined(FEAT_QUICKFIX)
-    win_T	*pvwin;
+    win_T   *pvwin;
 #endif
 
 #ifdef FEAT_RIGHTLEFT
-    pum_rl = State != MODE_CMDLINE && curwin->w_p_rl;
+    pum_rl = (State & MODE_CMDLINE) == 0 && curwin->w_p_rl;
 #endif
 
     do
     {
-	def_width = p_pw;
-	if (p_pmw > 0 && def_width > p_pmw)
-	    def_width = p_pmw;
 	above_row = 0;
 	below_row = cmdline_row;
 
@@ -129,7 +319,7 @@ pum_display(
 	// Remember the essential parts of the window position and size, so we
 	// can decide when to reposition the popup menu.
 	pum_window = curwin;
-	if (State == MODE_CMDLINE)
+	if (State & MODE_CMDLINE)
 	    // cmdline completion popup menu
 	    pum_win_row = cmdline_row;
 	else
@@ -152,87 +342,19 @@ pum_display(
 	}
 #endif
 
-	/*
-	 * Figure out the size and position of the pum.
-	 */
-	pum_height = MIN(size, PUM_DEF_HEIGHT);
-	if (p_ph > 0 && pum_height > p_ph)
-	    pum_height = p_ph;
-
-	// Put the pum below "pum_win_row" if possible.  If there are few lines
-	// decide on where there is more room.
-	if (pum_win_row + 2 >= below_row - pum_height
-		      && pum_win_row - above_row > (below_row - above_row) / 2)
-	{
-	    // pum above "pum_win_row"
-
-	    if (State == MODE_CMDLINE)
-		// for cmdline pum, no need for context lines
-		context_lines = 0;
-	    else
-		// Leave two lines of context if possible
-		context_lines = MIN(2, curwin->w_wrow - curwin->w_cline_row);
-
-	    if (pum_win_row >= size + context_lines)
-	    {
-		pum_row = pum_win_row - size - context_lines;
-		pum_height = size;
-	    }
-	    else
-	    {
-		pum_row = 0;
-		pum_height = pum_win_row - context_lines;
-	    }
-	    if (p_ph > 0 && pum_height > p_ph)
-	    {
-		pum_row += pum_height - p_ph;
-		pum_height = p_ph;
-	    }
-	}
-	else
-	{
-	    // pum below "pum_win_row"
-
-	    if (State == MODE_CMDLINE)
-		// for cmdline pum, no need for context lines
-		context_lines = 0;
-	    else
-	    {
-		// Leave two lines of context if possible
-		validate_cheight();
-		cline_visible_offset = curwin->w_cline_row +
-				    curwin->w_cline_height - curwin->w_wrow;
-		context_lines = MIN(3, cline_visible_offset);
-	    }
-
-	    pum_row = pum_win_row + context_lines;
-	    pum_height = MIN(below_row - pum_row, size);
-	    if (p_ph > 0 && pum_height > p_ph)
-		pum_height = p_ph;
-	}
+	// Figure out the vertical size and position of the pum.
+	pum_compute_vertical_placement(size, above_row, below_row);
 
 	// don't display when we only have room for one line
 	if (pum_height < 1 || (pum_height == 1 && size > 1))
 	    return;
 
-#if defined(FEAT_QUICKFIX)
-	// If there is a preview window above avoid drawing over it.
-	if (pvwin != NULL && pum_row < above_row && pum_height > above_row)
-	{
-	    pum_row = above_row;
-	    pum_height = pum_win_row - above_row;
-	}
-#endif
-
 	pum_array = array;
 	pum_size = size;
 	pum_compute_size();
-	max_width = pum_base_width;
-	if (p_pmw > 0 && max_width > p_pmw)
-	    max_width = p_pmw;
 
-	// Calculate column
-	if (State == MODE_CMDLINE)
+	// Calculate cursor column
+	if (State & MODE_CMDLINE)
 	    // cmdline completion popup menu
 	    cursor_col = cmdline_compl_startcol();
 	else
@@ -248,134 +370,10 @@ pum_display(
 	}
 
 	// if there are more items than room we need a scrollbar
-	if (pum_height < size)
-	{
-	    pum_scrollbar = 1;
-	    ++max_width;
-	}
-	else
-	    pum_scrollbar = 0;
+	pum_scrollbar = (pum_height < size) ? 1 : 0;
 
-	if (def_width < max_width)
-	    def_width = max_width;
-
-	if (((cursor_col < Columns - p_pw || cursor_col < Columns - max_width)
-#ifdef FEAT_RIGHTLEFT
-		    && !pum_rl)
-	       || (pum_rl && (cursor_col > p_pw || cursor_col > max_width)
-#endif
-	   ))
-	{
-	    // align pum with "cursor_col"
-	    pum_col = cursor_col;
-
-	    // start with the maximum space available
-#ifdef FEAT_RIGHTLEFT
-	    if (pum_rl)
-		pum_width = pum_col - pum_scrollbar + 1;
-	    else
-#endif
-		pum_width = Columns - pum_col - pum_scrollbar;
-
-	    content_width = max_width + pum_kind_width + pum_extra_width + 1;
-	    if (pum_width > content_width && pum_width > p_pw)
-	    {
-		// Reduce width to fit item
-		pum_width = MAX(content_width, p_pw);
-		if (p_pmw > 0 && pum_width > p_pmw)
-		    pum_width = p_pmw;
-	    }
-	    else if (((cursor_col > p_pw || cursor_col > max_width)
-#ifdef FEAT_RIGHTLEFT
-			&& !pum_rl)
-		|| (pum_rl && (cursor_col < Columns - p_pw
-			|| cursor_col < Columns - max_width)
-#endif
-		    ))
-	    {
-		// align pum edge with "cursor_col"
-#ifdef FEAT_RIGHTLEFT
-		if (pum_rl
-			&& W_ENDCOL(curwin) < max_width + pum_scrollbar + 1)
-		{
-		    pum_col = cursor_col + max_width + pum_scrollbar + 1;
-		    if (pum_col >= Columns)
-			pum_col = Columns - 1;
-		}
-		else if (!pum_rl)
-#endif
-		{
-		    right_edge_col = Columns - max_width - pum_scrollbar;
-		    if (curwin->w_wincol > right_edge_col && max_width <= p_pw)
-			// use full width to end of the screen
-			pum_col = MAX(0, right_edge_col);
-		}
-
-#ifdef FEAT_RIGHTLEFT
-		if (pum_rl)
-		    pum_width = pum_col - pum_scrollbar + 1;
-		else
-#endif
-		    pum_width = Columns - pum_col - pum_scrollbar;
-
-		if (pum_width < p_pw)
-		{
-		    pum_width = p_pw;
-		    if (p_pmw > 0 && pum_width > p_pmw)
-			pum_width = p_pmw;
-#ifdef FEAT_RIGHTLEFT
-		    if (pum_rl)
-		    {
-			if (pum_width > pum_col)
-			    pum_width = pum_col;
-		    }
-		    else
-#endif
-		    {
-			if (pum_width >= Columns - pum_col)
-			    pum_width = Columns - pum_col - 1;
-		    }
-		}
-		else if (pum_width > content_width && pum_width > p_pw)
-		{
-		    pum_width = MAX(content_width, p_pw);
-		    if (p_pmw > 0 && pum_width > p_pmw)
-			pum_width = p_pmw;
-		}
-		else if (p_pmw > 0 && pum_width > p_pmw)
-		{
-		    pum_width = p_pmw;
-		}
-	    }
-
-	}
-	else if (Columns < def_width)
-	{
-	    // not enough room, will use what we have
-#ifdef FEAT_RIGHTLEFT
-	    if (pum_rl)
-		pum_col = Columns - 1;
-	    else
-#endif
-		pum_col = 0;
-	    pum_width = Columns - 1;
-	    if (p_pmw > 0 && pum_width > p_pmw)
-		pum_width = p_pmw;
-	}
-	else
-	{
-	    if (max_width > p_pw)
-		max_width = p_pw;	// truncate
-	    if (p_pmw > 0 && max_width > p_pmw)
-		max_width = p_pmw;
-#ifdef FEAT_RIGHTLEFT
-	    if (pum_rl)
-		pum_col = max_width - 1;
-	    else
-#endif
-		pum_col = Columns - max_width;
-	    pum_width = max_width - pum_scrollbar;
-	}
+	// Figure out the horizontal size and position of the pum.
+	pum_compute_horizontal_placement(cursor_col);
 
 	// Set selected item and redraw.  If the window size changed need to
 	// redo the positioning.  Limit this to two times, when there is not
@@ -442,7 +440,7 @@ pum_compute_text_attrs(char_u *text, hlf_T hlf, int user_hlattr)
 	return NULL;
 
     is_select = hlf == HLF_PSI;
-    leader = State == MODE_CMDLINE ? cmdline_compl_pattern()
+    leader = (State & MODE_CMDLINE) ? cmdline_compl_pattern()
 							  : ins_compl_leader();
     if (leader == NULL || *leader == NUL)
 	return NULL;
@@ -451,12 +449,19 @@ pum_compute_text_attrs(char_u *text, hlf_T hlf, int user_hlattr)
     if (attrs == NULL)
 	return NULL;
 
-    in_fuzzy = State == MODE_CMDLINE ? cmdline_compl_is_fuzzy()
+    in_fuzzy = (State & MODE_CMDLINE) ? cmdline_compl_is_fuzzy()
 					  : (get_cot_flags() & COT_FUZZY) != 0;
     leader_len = STRLEN(leader);
 
     if (in_fuzzy)
+    {
 	ga = fuzzy_match_str_with_pos(text, leader);
+	if (!ga)
+	{
+	    vim_free(attrs);
+	    return NULL;
+	}
+    }
 
     while (*ptr != NUL)
     {
@@ -541,16 +546,6 @@ pum_screen_puts_with_attrs(
     }
 }
 
-
-    static inline void
-pum_align_order(int *order)
-{
-    int is_default = cia_flags == 0;
-    order[0] = is_default ? CPT_ABBR : cia_flags / 100;
-    order[1] = is_default ? CPT_KIND : (cia_flags / 10) % 10;
-    order[2] = is_default ? CPT_MENU : cia_flags % 10;
-}
-
     static inline char_u *
 pum_get_item(int index, int type)
 {
@@ -614,7 +609,7 @@ pum_display_rtl_text(
 
     char_u *rt_start = rt;
     cells = mb_string2cells(rt, -1);
-    truncated = width_limit == p_pmw && width_limit - totwidth < cells + pad;
+    truncated = width_limit - totwidth < cells + pad;
 
     // only draw the text that fits
     if (cells > width_limit)
@@ -658,25 +653,22 @@ pum_display_rtl_text(
 	width = cells + over_cell + 1;
 	rt = orig_rt;
 
-	screen_putchar(truncrl, row,
-		col - width + 1 + TPL_LCOL(NULL), trunc_attr);
+	screen_putchar(truncrl, row, col - width + 1, trunc_attr);
 
 	if (over_cell > 0)
-	    screen_fill(row, row + 1, col - width + 2 + TPL_LCOL(NULL),
-		    col - width + 2 + over_cell + TPL_LCOL(NULL), ' ', ' ',
-		    attr);
+	    screen_fill(row, row + 1, col - width + 2,
+		    col - width + 2 + over_cell, ' ', ' ', attr);
     }
 
     if (attrs == NULL)
-	screen_puts_len(rt, (int)STRLEN(rt), row,
-		col - cells + 1 + TPL_LCOL(NULL), attr);
+	screen_puts_len(rt, (int)STRLEN(rt), row, col - cells + 1, attr);
     else
-	pum_screen_puts_with_attrs(row, col - cells + 1 + TPL_LCOL(NULL),
-		cells, rt, (int)STRLEN(rt), attrs);
+	pum_screen_puts_with_attrs(row, col - cells + 1, cells, rt,
+		(int)STRLEN(rt), attrs);
 
     vim_free(rt_start);
     VIM_CLEAR(st);
-    return col - width + TPL_LCOL(NULL);
+    return col - width;
 }
 #endif
 
@@ -713,7 +705,7 @@ pum_display_ltr_text(
 
     size = (int)STRLEN(st);
     cells = (*mb_string2cells)(st, size);
-    truncated = width_limit == p_pmw && width_limit - totwidth < cells + pad;
+    truncated = width_limit - totwidth < cells + pad;
 
     // only draw the text that fits
     while (size > 0 && col + cells > width_limit + pum_col)
@@ -750,19 +742,17 @@ pum_display_ltr_text(
     }
 
     if (attrs == NULL)
-	screen_puts_len(st, size, row, col + TPL_LCOL(NULL), attr);
+	screen_puts_len(st, size, row, col, attr);
     else
-	pum_screen_puts_with_attrs(row, col + TPL_LCOL(NULL), cells, st, size,
-		attrs);
+	pum_screen_puts_with_attrs(row, col, cells, st, size, attrs);
 
     if (truncated)
     {
 	if (over_cell > 0)
-	    screen_fill(row, row + 1, col + cells + TPL_LCOL(NULL),
-		    col + cells + over_cell + TPL_LCOL(NULL), ' ', ' ', attr);
+	    screen_fill(row, row + 1, col + cells,
+		    col + cells + over_cell, ' ', ' ', attr);
 
-	screen_putchar(trunc, row,
-		col + cells + over_cell + TPL_LCOL(NULL), trunc_attr);
+	screen_putchar(trunc, row, col + cells + over_cell, trunc_attr);
     }
 
     VIM_CLEAR(st);
@@ -873,10 +863,19 @@ pum_draw_scrollbar(
 
 #ifdef FEAT_RIGHTLEFT
     if (pum_rl)
-	screen_putchar(' ', row, pum_col - pum_width + TPL_LCOL(NULL), attr);
+	screen_putchar(' ', row, pum_col - pum_width, attr);
     else
 #endif
-	screen_putchar(' ', row, pum_col + pum_width + TPL_LCOL(NULL), attr);
+	screen_putchar(' ', row, pum_col + pum_width, attr);
+}
+
+    static inline void
+pum_align_order(int *order)
+{
+    int is_default = cia_flags == 0;
+    order[0] = is_default ? CPT_ABBR : cia_flags / 100;
+    order[1] = is_default ? CPT_KIND : (cia_flags / 10) % 10;
+    order[2] = is_default ? CPT_MENU : cia_flags % 10;
 }
 
 /*
@@ -947,6 +946,12 @@ pum_redraw(void)
     screen_zindex = POPUPMENU_ZINDEX;
 #endif
 
+    // Draw border and shadow first if enabled
+    if (pum_border)
+	pum_draw_border();
+    if (pum_shadow)
+	pum_draw_shadow();
+
     for (i = 0; i < pum_height; ++i)
     {
 	idx = i + pum_first;
@@ -958,13 +963,13 @@ pum_redraw(void)
 #ifdef FEAT_RIGHTLEFT
 	if (pum_rl)
 	{
-	    if (pum_col < curwin->w_wincol + curwin->w_width - 1)
-		screen_putchar(' ', row, pum_col + 1 + TPL_LCOL(NULL), attr);
+	    if (pum_col < curwin->w_wincol + curwin->w_width - 1 - pum_border)
+		screen_putchar(' ', row, pum_col + 1, attr);
 	}
 	else
 #endif
-	    if (pum_col > 0)
-		screen_putchar(' ', row, pum_col - 1 + TPL_LCOL(NULL), attr);
+	    if (pum_col > pum_border)
+		screen_putchar(' ', row, pum_col - 1, attr);
 
 	// Display each entry, use two spaces for a Tab.
 	// Do this 3 times and order from p_cia
@@ -985,6 +990,8 @@ pum_redraw(void)
 
 	    if (j + 1 < 3)
 		next_isempty = pum_get_item(idx, order[j + 1]) == NULL;
+	    else
+		next_isempty = TRUE;
 
 	    if (p != NULL)
 		// Process and display the item
@@ -1005,16 +1012,15 @@ pum_redraw(void)
 #ifdef FEAT_RIGHTLEFT
 	    if (pum_rl)
 	    {
-		screen_fill(row, row + 1, pum_col - basic_width - n + 1 + TPL_LCOL(NULL),
-						col + 1 + TPL_LCOL(NULL), ' ', ' ', orig_attr);
+		screen_fill(row, row + 1, pum_col - basic_width - n + 1,
+						col + 1, ' ', ' ', orig_attr);
 		col = pum_col - basic_width - n;
 	    }
 	    else
 #endif
 	    {
-		screen_fill(row, row + 1, col + TPL_LCOL(NULL),
-			pum_col + basic_width + n + TPL_LCOL(NULL), ' ', ' ',
-			orig_attr);
+		screen_fill(row, row + 1, col, pum_col + basic_width + n,
+							' ', ' ', orig_attr);
 		col = pum_col + basic_width + n;
 	    }
 	    totwidth = basic_width + n;
@@ -1022,14 +1028,12 @@ pum_redraw(void)
 
 #ifdef FEAT_RIGHTLEFT
 	if (pum_rl)
-	    screen_fill(row, row + 1,
-		    pum_col - pum_width + 1 + TPL_LCOL(NULL),
-		    col + 1 + TPL_LCOL(NULL), ' ', ' ', orig_attr);
+	    screen_fill(row, row + 1, pum_col - pum_width + 1, col + 1, ' ',
+							    ' ', orig_attr);
 	else
 #endif
-	    screen_fill(row, row + 1, col + TPL_LCOL(NULL),
-		    pum_col + pum_width + TPL_LCOL(NULL),
-		    ' ', ' ', orig_attr);
+	    screen_fill(row, row + 1, col, pum_col + pum_width, ' ', ' ',
+								orig_attr);
 	pum_draw_scrollbar(row, i, thumb_pos, thumb_height);
 
 	++row;
@@ -1040,7 +1044,7 @@ pum_redraw(void)
 #endif
 }
 
-#if (defined(FEAT_PROP_POPUP) && defined(FEAT_QUICKFIX)) || defined(PROTO)
+#if defined(FEAT_PROP_POPUP) && defined(FEAT_QUICKFIX)
 /*
  * Position the info popup relative to the popup menu item.
  */
@@ -1051,14 +1055,22 @@ pum_position_info_popup(win_T *wp)
     int row = pum_row;
     int botpos = POPPOS_BOTLEFT;
     int	used_maxwidth_opt = FALSE;
+    int	right_margin, left_margin, dummy;
+    int left_extra_width, right_extra_width;
+
+    (void)compute_margins(&right_margin, &left_margin, &dummy);
+
+    left_extra_width = pum_border + left_margin;
+    right_extra_width = pum_border + right_margin + (pum_shadow ? 2 : 0);
+    col += right_extra_width;
 
     wp->w_popup_pos = POPPOS_TOPLEFT;
     if (Columns - col < 20 && Columns - col < pum_col)
     {
-	col = pum_col - 1;
+	col = pum_col - 1 - left_extra_width;
 	wp->w_popup_pos = POPPOS_TOPRIGHT;
 	botpos = POPPOS_BOTRIGHT;
-	wp->w_maxwidth = pum_col - 1;
+	wp->w_maxwidth = pum_col - 1 - left_extra_width;
     }
     else
 	wp->w_maxwidth = Columns - col + 1;
@@ -1169,12 +1181,14 @@ pum_set_selected(int n, int repeat UNUSED)
 	 * 'completeopt' contains "preview" or "popup" or "popuphidden".
 	 * Skip this when tried twice already.
 	 * Skip this also when there is not much room.
+	 * Skip this for command-window when 'completeopt' contains "preview".
 	 * NOTE: Be very careful not to sync undo!
 	 */
 	if (pum_array[pum_selected].pum_info != NULL
 		&& Rows > 10
 		&& repeat <= 1
-		&& (cur_cot_flags & COT_ANY_PREVIEW))
+		&& (cur_cot_flags & COT_ANY_PREVIEW)
+		&& !((cur_cot_flags & COT_PREVIEW) && cmdwin_type != 0))
 	{
 	    win_T	*curwin_save = curwin;
 	    tabpage_T   *curtab_save = curtab;
@@ -1513,7 +1527,7 @@ pum_get_height(void)
     return pum_height;
 }
 
-#if defined(FEAT_EVAL) || defined(PROTO)
+#if defined(FEAT_EVAL)
 /*
  * Add size information about the pum to "dict".
  */
@@ -1532,54 +1546,67 @@ pum_set_event_info(dict_T *dict)
 }
 #endif
 
-#if defined(FEAT_BEVAL_TERM) || defined(FEAT_TERM_POPUP_MENU) || defined(PROTO)
+#if defined(FEAT_BEVAL_TERM) || defined(FEAT_TERM_POPUP_MENU)
     static void
 pum_position_at_mouse(int min_width)
 {
-    if (Rows - mouse_row > pum_size || Rows - mouse_row > mouse_row)
+    int extra_height = pum_border * 2 + pum_shadow;
+    int extra_offset = pum_border + pum_shadow;
+    int trailing_overhead = 0;
+    int col_offset = 0;
+    int available_width;
+
+    if (Rows - mouse_row - extra_height > pum_size
+	    || Rows - mouse_row > mouse_row)
     {
 	// Enough space below the mouse row,
 	// or there is more space below the mouse row than above.
-	pum_row = mouse_row + 1;
-	if (pum_height > Rows - pum_row)
-	    pum_height = Rows - pum_row;
-	if (pum_row + pum_height > cmdline_row)
+	pum_row = mouse_row + 1 + pum_border;
+	if (pum_height > Rows - pum_row - extra_height)
+	    pum_height = Rows - pum_row - extra_height;
+	if (pum_row + pum_height + extra_offset >= cmdline_row)
 	    pum_in_cmdline = TRUE;
     }
     else
     {
 	// Show above the mouse row, reduce height if it does not fit.
-	pum_row = mouse_row - pum_size;
-	if (pum_row < 0)
+	pum_row = mouse_row - pum_size - extra_offset;
+	if (pum_row < pum_border)
 	{
-	    pum_height += pum_row;
-	    pum_row = 0;
+	    pum_height += pum_row - pum_border;
+	    pum_row = pum_border;
 	}
     }
 
+    trailing_overhead = pum_scrollbar + pum_border
+	+ (pum_margin && pum_border ? 1 : 0) + (pum_shadow ? 2 : 0);
+
 # ifdef FEAT_RIGHTLEFT
     if (pum_rl)
     {
-	if (mouse_col + 1 >= pum_base_width
-		|| mouse_col + 1 > min_width)
+	col_offset = (mouse_col == Columns - 1) && pum_border;
+	available_width = mouse_col + 1 - trailing_overhead - col_offset;
+	if (available_width >= pum_base_width || available_width > min_width)
 	    // Enough space to show at mouse column.
-	    pum_col = mouse_col;
+	    pum_col = mouse_col - col_offset;
 	else
 	    // Not enough space, left align with window.
-	    pum_col = MIN(pum_base_width, min_width) - 1;
-	pum_width = pum_col + 1;
+	    pum_col = MIN(pum_base_width, min_width) - 1 + trailing_overhead;
+	pum_width = pum_col + 1 - trailing_overhead;
     }
     else
 # endif
     {
-	if (Columns - mouse_col >= pum_base_width
-		|| Columns - mouse_col > min_width)
+	col_offset = (mouse_col == 0) && pum_border;
+	available_width = Columns - mouse_col - trailing_overhead - col_offset;
+	if (available_width >= pum_base_width || available_width > min_width)
 	    // Enough space to show at mouse column.
-	    pum_col = mouse_col;
+	    pum_col = mouse_col + col_offset;
 	else
 	    // Not enough space, right align with window.
-	    pum_col = Columns -  MIN(pum_base_width, min_width);
-	pum_width = Columns - pum_col;
+	    pum_col = Columns - MIN(pum_base_width, min_width)
+		- trailing_overhead;
+	pum_width = Columns - pum_col - trailing_overhead;
     }
 
     pum_width = MIN(pum_width, pum_base_width + 1);
@@ -1589,7 +1616,7 @@ pum_position_at_mouse(int min_width)
 
 #endif
 
-#if defined(FEAT_BEVAL_TERM) || defined(PROTO)
+#if defined(FEAT_BEVAL_TERM)
 static pumitem_T *balloon_array = NULL;
 static int balloon_arraysize;
 
@@ -1825,7 +1852,7 @@ ui_may_remove_balloon(void)
 }
 #endif
 
-#if defined(FEAT_TERM_POPUP_MENU) || defined(PROTO)
+#if defined(FEAT_TERM_POPUP_MENU)
 /*
  * Select the pum entry at the mouse position.
  */
@@ -2026,3 +2053,250 @@ pum_make_popup(char_u *path_name, int use_mouse_pos)
 	pum_show_popupmenu(menu);
 }
 #endif
+
+    static int
+is_singlewidth(int c)
+{
+    if (c >= 0 && c < 0x80)
+	return TRUE;
+    if (utf_char2cells(c) == 1)
+	return TRUE;
+    return FALSE;
+}
+
+/*
+ * Set border characters for popup menu.
+ */
+    void
+pum_set_border_chars(
+	int top,
+	int right,
+	int bottom,
+	int left,
+	int top_left,
+	int top_right,
+	int bottom_right,
+	int bottom_left)
+{
+    if (!is_singlewidth(top) || !is_singlewidth(right)
+	    || !is_singlewidth(bottom) || !is_singlewidth(left)
+	    || !is_singlewidth(top_left) || !is_singlewidth(top_right)
+	    || !is_singlewidth(bottom_right) || !is_singlewidth(bottom_left))
+	return;
+    pum_border_chars.top = top;
+    pum_border_chars.right = right;
+    pum_border_chars.bottom = bottom;
+    pum_border_chars.left = left;
+    pum_border_chars.top_left = top_left;
+    pum_border_chars.top_right = top_right;
+    pum_border_chars.bottom_right = bottom_right;
+    pum_border_chars.bottom_left = bottom_left;
+    pum_border = 1;
+}
+
+/*
+ * Compute margins for the popup menu.
+ */
+    static void
+compute_margins(int *right_margin, int *left_margin, int *left_padding)
+{
+    *right_margin = pum_margin && pum_border;
+
+#ifdef FEAT_RIGHTLEFT
+    if (pum_rl)
+    {
+	*left_padding = (pum_col < Columns - 1 - pum_border);
+	*left_margin = *right_margin && (pum_col < Columns - 3);
+	return;
+    }
+#endif
+    *left_padding = (pum_col > pum_border);
+    *left_margin = *right_margin && (pum_col > 2);
+}
+
+/*
+ * Draw the border around the popup menu
+ */
+    static void
+pum_draw_border(void)
+{
+    int	top_row = pum_row - 1;
+    int	bottom_row = pum_row + pum_height;
+    int	col;
+    int	attr = highlight_attr[HLF_PMB];
+    int	right_margin, left_margin, left_padding;
+    int	inner_width;
+
+    if (!pum_border)
+	return;
+
+    (void)compute_margins(&right_margin, &left_margin, &left_padding);
+
+    inner_width = pum_width + pum_scrollbar + left_padding;
+
+    // Draw border left to right or right to left (for RTL)
+#ifdef FEAT_RIGHTLEFT
+    if (pum_rl)
+	col = pum_col + left_padding + 1;
+    else
+#endif
+	col = pum_col - left_padding - 1;
+
+    // left corners (right corners for RTL)
+#ifdef FEAT_RIGHTLEFT
+    if (pum_rl)
+    {
+	screen_putchar(pum_border_chars.top_right, top_row, col, attr);
+	if (left_margin)
+	    screen_putchar(' ', top_row, col + 1, attr);
+	screen_putchar(pum_border_chars.bottom_right, bottom_row, col, attr);
+	if (left_margin)
+	    screen_putchar(' ', bottom_row, col + 1, attr);
+    }
+    else
+#endif
+    {
+	screen_putchar(pum_border_chars.top_left, top_row, col, attr);
+	if (left_margin)
+	    screen_putchar(' ', top_row, col - 1, attr);
+	screen_putchar(pum_border_chars.bottom_left, bottom_row, col, attr);
+	if (left_margin)
+	    screen_putchar(' ', bottom_row, col - 1, attr);
+    }
+
+    // Top horizontal lines
+    for (int i = 0; i < inner_width; i++)
+    {
+	int cur_col;
+#ifdef FEAT_RIGHTLEFT
+	if (pum_rl)
+	    cur_col = col - 1 - i;
+	else
+#endif
+	    cur_col = col + 1 + i;
+	screen_putchar(pum_border_chars.top, top_row, cur_col, attr);
+	screen_putchar(pum_border_chars.bottom, bottom_row, cur_col, attr);
+    }
+
+    // right corners (left corners for RTL)
+#ifdef FEAT_RIGHTLEFT
+    if (pum_rl)
+    {
+	screen_putchar(pum_border_chars.top_left, top_row,
+		col - inner_width - 1, attr);
+	if (right_margin)
+	    screen_putchar(' ', top_row, col - inner_width - 2, attr);
+	screen_putchar(pum_border_chars.bottom_left, bottom_row,
+		col - inner_width - 1, attr);
+	if (right_margin)
+	    screen_putchar(' ', bottom_row, col - inner_width - 2, attr);
+    }
+    else
+#endif
+    {
+	screen_putchar(pum_border_chars.top_right, top_row,
+		col + inner_width + 1, attr);
+	if (right_margin)
+	    screen_putchar(' ', top_row, col + inner_width + 2, attr);
+	screen_putchar(pum_border_chars.bottom_right, bottom_row,
+		col + inner_width + 1, attr);
+	if (right_margin)
+	    screen_putchar(' ', bottom_row, col + inner_width + 2, attr);
+    }
+
+    // Draw side borders
+    for (int i = 0; i < pum_height; i++)
+    {
+	int row = pum_row + i;
+#ifdef FEAT_RIGHTLEFT
+	if (pum_rl)
+	{
+	    screen_putchar(pum_border_chars.left, row, col, attr);
+	    if (left_margin)
+		screen_putchar(' ', row, col + 1, attr);
+	    screen_putchar(pum_border_chars.right, row,
+		    col - inner_width - 1, attr);
+	    if (right_margin)
+		screen_putchar(' ', row, col - inner_width - 2, attr);
+	}
+	else
+#endif
+	{
+	    screen_putchar(pum_border_chars.left, row, col, attr);
+	    if (left_margin)
+		screen_putchar(' ', row, col - 1, attr);
+	    screen_putchar(pum_border_chars.right, row,
+		    col + inner_width + 1, attr);
+	    if (right_margin)
+		screen_putchar(' ', row, col + inner_width + 2, attr);
+	}
+    }
+}
+
+/*
+ * Get the underlying character and redraw with shadow highlight
+ */
+    void
+put_shadow_char(int row, int col)
+{
+    char_u  buf[MB_MAXBYTES + 1];
+    int	    attr = highlight_attr[HLF_PMS];
+
+    screen_getbytes(row, col, buf, NULL);
+    screen_putchar((*mb_ptr2char)(buf), row, col, attr);
+}
+
+/*
+ * Draw the shadow
+ */
+    static void
+pum_draw_shadow(void)
+{
+    int	row, col, next_col;
+    int	width, offset;
+    int	inner_width;
+    int	right_margin, left_margin, left_padding;
+
+    if (!pum_shadow)
+	return;
+
+    (void)compute_margins(&right_margin, &left_margin, &left_padding);
+    inner_width = pum_width + pum_scrollbar + pum_border + right_margin;
+
+    // Draw right side shadow (left for RTL)
+#ifdef FEAT_RIGHTLEFT
+    if (pum_rl)
+    {
+	col = pum_col - inner_width;
+	next_col = col - 1;
+    }
+    else
+#endif
+    {
+	col = pum_col + inner_width;
+	next_col = col + 1;
+    }
+
+    for (int i = 0; i < pum_height + pum_border; i++)
+    {
+	row = pum_row + 1 + i - pum_border;
+	put_shadow_char(row, col);
+	put_shadow_char(row, next_col);
+    }
+
+    // Bottom shadow
+    row = pum_row + pum_height + pum_border;
+    width = inner_width + left_padding + pum_border + left_margin;
+    offset = 2 - left_padding - pum_border - left_margin;
+
+    for (int i = 0; i < width; i++)
+    {
+#ifdef FEAT_RIGHTLEFT
+	if (pum_rl)
+	    col = pum_col - offset - i;
+	else
+#endif
+	    col = pum_col + offset + i;
+	put_shadow_char(row, col);
+    }
+}
